#include "Parse.h"

int Message_length[NUM_MESSAGE] =              /* Expected bytes of each received message */
{
	87,  /* Number of bytes in Position/Time/Status Message      */
	80   /* Number of bytes in Visible Satellites Status Message */
};

//VSS_DATA       Vss_data;                      /* Visible satellites status data     */
//POSITION_DATA  Position_data;                 /* Position data                      */

int            InBufIdx = 0;                  /* Index of InBuf                     */
unsigned char  InBuf[IN_BUF_LEN];             /* Buffer to save streams             */

/* Global functions */



/* Local  functions */
void  NBIP_init_data(void);
int   get_message_type(int index);
void  parse_PTSR(unsigned char* buf, int len);
void  parse_VSSMR(unsigned char* buf);
long  convert_value(unsigned char* msg, int start, int end);
short convert_int(unsigned char* msg, int start, int end);


int NBIP_parse_buf(unsigned char *buf, int len)
{
	
	int i, j, k = 0;
	int retval;
	int message_type;
	int check_sum_index;
	int message_index=0, end_of_message=0;
	int complete_flag = 1;
	unsigned char check_sum = 0, message[MESSAGE_LEN];

	/* Initialize return value */
	retval = NO_MESSAGE;

	/* Append input buffer to InBuf */
	for(i = 0; i < len; i++) {
		InBuf[InBufIdx++] = buf[i];
	}

	/* Check if message is completed and parse into the structure */
	for(i = 0; i < InBufIdx; i++) {

		/* If the start mark("::") detected      */
		if(InBuf[i] == ':' && InBuf[i+1] == ':') {

			/* If "::" mark is at the end of buf, break 
			   and append the start mark to the start of InBuf */
			if(i+2 >= InBufIdx - 1 ) {	
				complete_flag = FALSE;
				break;
			}

			/* Get message type */
			message_type = get_message_type(i+2);

			/* If invalid message type detected,
			   ignore three bytes(::<message type>) and continue with remainder in buf */
			if(message_type == NO_MESSAGE) {
				i += 3;
				continue;
			}

			/* If received message length is shorter than the expected message length,
			   continue with remainder in buf */
			if(i + Message_length[message_type] > InBufIdx) {
				complete_flag = FALSE;
				break;
			}


			/* Get checksum index */
			check_sum_index = i + Message_length[message_type]-3;

			/* Calculate checksum */
			for(j = i + 2; j < check_sum_index; j++) {
				check_sum ^= InBuf[j];
			}

			/* If checksum error detected, ignore two bytes and continue with remainder in buf */
			if(check_sum != InBuf[check_sum_index]) { 
				check_sum = 0;
				i += Message_length[message_type]-1;
				continue;
			}

			/* Save valid message to sentence buffer */
			message_index = 0;
			for(end_of_message = i; end_of_message < i+Message_length[message_type]; end_of_message++) {
					message[message_index++] = InBuf[end_of_message];
			}

			/* Parse stream and save to proper structure */
			switch(message_type) {

				case Position_Time_Status_Request :
					NBIP_init_data();
					parse_PTSR(message, message_index);
					retval = Position_Time_Status_Request;
					break;

				case Visible_Satellites_Status_Messge_Request :
					NBIP_init_data();
					parse_VSSMR(message);
					retval = Visible_Satellites_Status_Messge_Request;
					break;

				default :
					retval = NO_MESSAGE;
			}
			i += Message_length[message_type]-1;
			check_sum = 0;

		} /* End of if */
	} /* End of for */


	/* If the end of buf is the start character of following new message(e.g. ':'), 
	   put it(':') into InBuf[0] and set InBufIdx to 1 */
	if(InBuf[InBufIdx-3] == 0x0D && InBuf[InBufIdx-2] == 0x0A && InBuf[InBufIdx-1] == ':') {
		InBuf[0] = InBuf[InBufIdx-1]; 
		InBufIdx = 1;
		return NO_MESSAGE;
	}

	k = 0;

	/* If a message is parsed and a new message is still not completely constructed,
	   shift them to left-most part of InBuf */
	if(complete_flag == FALSE) {
		for( j = i; j < InBufIdx; j++ ) {
			InBuf[k++] = InBuf[j] ;
		}
		InBufIdx = k ;
	} 

	/* if current message is parsed and there are remainders in InBuf,
	   shift them to left-most part of InBuf */
	else {
		if( InBufIdx > message_index ) {
			for( j = message_index; j < InBufIdx; j++) 
				InBuf[k++] = InBuf[j];
			InBufIdx = k;
		} 
		else
			InBufIdx = 0 ;
	}


	return retval;

}


int get_message_type(int index)
{
	int  message_type;

	switch(InBuf[index]) {
		
		case 'I' :
			switch(InBuf[index+1]) {
				case '6' :
					message_type = Position_Time_Status_Request;
					break;
				default :
					message_type = NO_MESSAGE;
					break;
			}
			break;

		case 'M' :
			switch(InBuf[index+1]) {
				case '5' :
					message_type = Visible_Satellites_Status_Messge_Request;
					break;
				default :
					message_type = NO_MESSAGE;
					break;
			}
			break;

		default :
			message_type = NO_MESSAGE;
			break;
	}

	return message_type;
}

void DoubleToChar(double value, unsigned char *result)
{
	char *ch;

	ch = (char*)&value;

	result[0] = (unsigned char)*(ch++);
	result[1] = (unsigned char)*(ch++);
	result[2] = (unsigned char)*(ch++);
	result[3] = (unsigned char)*(ch++);
	result[4] = (unsigned char)*(ch++);
	result[5] = (unsigned char)*(ch++);
	result[6] = (unsigned char)*(ch++);
	result[7] = (unsigned char)*(ch++);
}


void parse_PTSR(unsigned char* buf, int len)
{
	int    channel, index, temp = 0xff;
	
	Position_data.Time.Hour   = buf[4];
	Position_data.Time.Minute = buf[5];
	Position_data.Time.Second = buf[6];

	Position_data.Date.Day   = buf[7];
	Position_data.Date.Month = buf[8];
	Position_data.Date.Year  = buf[9];

	Position_data.Position.Latitude  = convert_value(buf, 10, 13);
	Position_data.Position.Longitude = convert_value(buf, 14, 17);
	Position_data.Position.Alt_gps   = convert_value(buf, 18, 21);
	Position_data.Position.Alt_msl   = convert_value(buf, 22, 25);

	Position_data.Velocity.G_velocity = (unsigned int)convert_value(buf, 26, 27);
	Position_data.Velocity.Track_angle = (int)convert_value(buf, 28, 29);

	Position_data.Dop.Type = buf[30];
	Position_data.Dop.Value = (int)convert_value(buf, 31, 32);

	Position_data.Track_SV = buf[33];
	Position_data.Visible_SV = buf[34];

	index = 35;

	for(channel = 0; channel < NUM_CHANNEL; channel++) {
		Position_data.Channel[channel].SV_ID      = buf[index++];
		Position_data.Channel[channel].Track_mode = buf[index++];

		temp = buf[index++];

		Position_data.Channel[channel].Pos_fix    = temp & 1;
		Position_data.Channel[channel].Unhealthy  = ( temp & 2  ) >> 1;
		Position_data.Channel[channel].Not_using_channel = ( temp & 4  ) >> 2;
																	
		Position_data.Channel[channel].Signal_strength = buf[index++];
	} 
	
	temp = buf[83];

	Position_data.Receiver.Fix_mode        = temp & 1;
	Position_data.Receiver.Poor_geo        = ( temp & 2  ) >> 1;
	Position_data.Receiver.ThreeD_fix      = ( temp & 4  ) >> 2;
	Position_data.Receiver.Alt_hold        = ( temp & 8  ) >> 3;
	Position_data.Receiver.Diff            = ( temp & 16 ) >> 4;
	Position_data.Receiver.Insufficient_VS = ( temp & 32 ) >> 5;
	Position_data.Receiver.Bad_almanac     = ( temp & 64 ) >> 6;

}



void parse_VSSMR(unsigned char* buf)
{
	int  sv, index = 4 ;

	/* Total number of visible satellites */
	Vss_data.Total_sat_num = buf[index++] ;

	for(sv = 0; sv < NUM_CHANNEL; sv++) {
		/* Each satellite's ID */
		Vss_data.Visible_sat[sv].Sat_id    = (int)(buf[index] & 0x1f) ;
		/* Each satellite's Healthy */
		Vss_data.Visible_sat[sv].Healthy   = (int)( (int)( buf[index++] & 0x20 ) >> 5 ) ;
		/* Each satellite's Doppler */
		Vss_data.Visible_sat[sv].Doppler   = convert_int(buf, index, index+1) ;
		index += 2 ;
		/* Each satellite's Elevation */
		Vss_data.Visible_sat[sv].Elevation = buf[index++] ;
		/* Each satellite's Azimuth */
		Vss_data.Visible_sat[sv].Azimuth   = (int)convert_value(buf, index, index+1) ;
		index += 2 ;
	}
}



void NBIP_init_data(void)
{
	int i;

	/* Initialize Position/Time/Status Data */	
	Position_data.Time.Hour   = 0;
	Position_data.Time.Minute = 0;
	Position_data.Time.Second = 0;
	Position_data.Date.Day    = 0;
	Position_data.Date.Month  = 0;
	Position_data.Date.Year   = 0;
	Position_data.Position.Latitude  = 0;
	Position_data.Position.Longitude = 0;
	Position_data.Position.Alt_gps   = 0;
	Position_data.Position.Alt_msl   = 0;
	Position_data.Velocity.G_velocity = 0;
	Position_data.Velocity.Track_angle = 0;
	Position_data.Dop.Type = 'X';
	Position_data.Dop.Value = 0;
	Position_data.Track_SV = 0;
	Position_data.Visible_SV = 0;
	for(i = 0; i < NUM_CHANNEL; i++) {
		Position_data.Channel[i].SV_ID      = 0;
		Position_data.Channel[i].Track_mode = 0;
		Position_data.Channel[i].Pos_fix    = 0;
		Position_data.Channel[i].Unhealthy  = 1;
		Position_data.Channel[i].Not_using_channel = 1;
		Position_data.Channel[i].Signal_strength = 0;
	} 
	Position_data.Receiver.Fix_mode        = 0;
	Position_data.Receiver.Poor_geo        = 0;
	Position_data.Receiver.ThreeD_fix      = 0;
	Position_data.Receiver.Alt_hold        = 0;
	Position_data.Receiver.Diff            = 0;
	Position_data.Receiver.Insufficient_VS = 0;
	Position_data.Receiver.Bad_almanac     = 0;

	/* Initialize Visible Satellite Status */
	Vss_data.Total_sat_num = 0;
	for(i = 0; i < NUM_CHANNEL; i++) {
		Vss_data.Visible_sat[i].Sat_id    = 0;
		Vss_data.Visible_sat[i].Healthy   = 0;
		Vss_data.Visible_sat[i].Doppler   = 0;
		Vss_data.Visible_sat[i].Elevation = 0;
		Vss_data.Visible_sat[i].Azimuth   = 0;
	}

}


long convert_value(unsigned char* msg, int start, int end)
{
	int  i;
	long temp = 0, j = 1;

	for(i = end; i >= start; i--) {
		temp += (long)(msg[i] * j);
		j    *= 256;
	}

	return temp;
}

short convert_int(unsigned char* msg, int start, int end)
{
	short i, temp = 0, j = 1;

	for(i = end; i >= start; i--) {
		temp += msg[i] * j;
		j    *= 256;
	}

	return temp;
}

unsigned long convert_ulong(unsigned char* msg, int start, int end)
{
	short         i;
	unsigned long temp = 0, j = 1;

	for(i = end; i >= start; i--) {
		temp += (unsigned long)(msg[i] * j);
		j    *= 256;
	}

	return temp;
}
